//-------------------------------------------------------------
//  中波 AM 放送用 SDR，CIC フィルタと通常の FIR フィルタ使用
//      選局をスイッチで行う方式
//
//      1/15 にダウンサンプリング：CIC フィルタ使用
//      1/4 にダウンサンプリング：通常の FIR フィルタ使用
//
//      入力ピン： A1
//      出力ピン： A2
//
//  2023/08/16, Copyright (c) 2023 MIKAMI, Naoki
//-------------------------------------------------------------

#include "F446_ADC_Intr.hpp"    // ADC
#include "F446_DAC.hpp"         // DAC
#include "QuadOscIir.hpp"       // 直交信号発生器
#include "Cic3Stage.hpp"        // ３段 CIC フィルタ
#include "CoefficientsLpf.hpp"  // FIR フィルタの係数
#include "FirSymmetry.hpp"      // 係数対称 FIR フィルタ
#include "Iir1st.hpp"           // 一次の IIR フィルタ
#include "AQM1602.hpp"          // LCD 表示器用
using namespace Mikami;
#pragma diag_suppress 870   // マルチバイト文字使用の警告抑制のため

const float FS_ = 900.0f;   // 標本化周波数，単位： kHz
const float T0_ = 1000/FS_; // 標本化間隔，単位：μs
const float A0_ = 8192;     // 直交信号発生器の出力の振幅

// 各放送局の搬送波の周波数，単位： Hz
const float F_C_[] = {  594.0e3f,   // NHK 第１
                        693.0e3f,   // NHK 第２
                        810.0e3f,   // AFN Tokyo
                        954.0e3f,   // TBS ラジオ
                       1134.0e3f,   // 文化放送
                       1242.0e3f,   // ニッポン放送
                       1422.0e3f};  // ラジオ日本

AdcF446_Intr adc_(FS_, A1);     // ADC オブジェクト
DacF446 dac_;                   // DAC オブジェクト
QuadOscIir nco_(0, T0_, A0_);   // NCO (numerically- controlled oscillator)
Iir1st averaging_(0.9998f);     // 直流分を求めるためのフィルタ

const int N1_ = 15;         // 1/15 にダウンサンプリング
Cic3Stage cic3I_(N1_, A0_); // CIC フィルタ，I 信号用
Cic3Stage cic3Q_(N1_, A0_); // CIC フィルタ，Q 信号用

const int N2_ = 4;          // 1/4 にダウンサンプリング
FirSymmetry lpfI_(ORDER_FIR_, HN_); // FIR フィルタ，I 信号用
FirSymmetry lpfQ_(ORDER_FIR_, HN_); // FIR フィルタ，Q 信号用

// AD 変換終了割り込みに対する割り込みサービス･ルーチン
void AdcIsr()
{
    static int count = 0;

    float xn = adc_.Read();     // ADC の読込み
    float sinX, cosX;
    nco_.Generate(sinX, cosX);  // sin, cos 発生

    // cos/sin を乗算したデータに対する CIC フィルタの積分処理
    cic3I_.Integrate((int16_t)(xn*cosX));   // I 信号用
    cic3Q_.Integrate((int16_t)(xn*sinX));   // Q 信号用

    if (++count < N1_) return;

    NVIC->STIR = EXTI3_IRQn;    // ソフトウェア割り込み発生
    count = 0;
}

// EXTI3_IRQn に対する割込みサービス･ルーチン
//      CIC フィルタの櫛形フィルタの処理と FIR フィルタの入力の処理
void IsrCIC()
{
    static int count = 0;

    // 櫛形フィルタの出力を FIR フィルタのバッファへ保存
    lpfI_.Store(cic3I_.CombFilter());   // I 信号用
    lpfQ_.Store(cic3Q_.CombFilter());   // Q 信号用

    if (++count < N2_) return;

    NVIC->STIR = EXTI4_IRQn;    // ソフトウェア割り込み発生
    count = 0;
}

// EXTI4_IRQn に対する割込みサービス･ルーチン
//      FIR フィルタの実行と AM 復調の処理
void IsrFIR()
{
    float yI = lpfI_.Execute(); // FIR フィルタによる LPF 実行
    float yQ = lpfQ_.Execute(); // FIR フィルタによる LPF 実行

    float yn = sqrtf(yI*yI + yQ*yQ);    // 包絡線成分を求める
    float dc = averaging_.Execute(yn);  // 直流分を求める
    yn = yn - dc;   // 直流分除去

    float gain = 0.8f/dc;               // 直流分から AGC の利得を計算
    if (gain > 200.0f) gain = 200.0f;   // gain の上限設定
    dac_.Write(gain*yn);                // AGC 後のデータを出力
}

// 選局の処理
int Tune(Aqm1602 &lcd, BusIn &sw)
{
    const static string RS[] = {
        "NHK #1", "NHK #2", "AFN", "TBS",
        "JOQR", "NBS", "RADIO NIPPON"};
    NVIC_DisableIRQ(EXTI4_IRQn);    // 復調処理一時停止
    wait(0.1f);                     // 実行待ち
    int k = sw.read();              // 再度 SW の値を入力
    lcd.WriteValueXY("SW: %d", k);  // SW の値を LCD に表示
    lcd.ClearLine(1);               // LCD の下の行の表示をクリア
    if (k>6) return -1;             // 対応する放送局なし
    lcd.WriteStringXY(RS[k], 0, 1); // 放送局のシンボルを LCD に表示
    printf("%d\r\n", k);            // SW の値を表示
    nco_.Set(F_C_[k]);              // NCO の周波数を変更し放送局を変更
    lpfI_.Clear();                  // FIR フィルタのバッファをクリア
    lpfQ_.Clear();                  // FIR フィルタのバッファをクリア
    NVIC_EnableIRQ(EXTI4_IRQn);     // 復調処理再開
    return k;                       // SW の値を戻り値とする
}

int main()
{
    Aqm1602 lcd;        // 液晶表示器，D14: SDA, D15: SCL
    BusIn sw(D10, D11, D12);    // ロータリ DIP スイッチ用
    sw.mode(PullDown);          // 入力：プルダウンに設定

    NVIC_SetPriority(ADC_IRQn,   0);    // 最優先
    NVIC_SetPriority(EXTI3_IRQn, 1);    // ２番目に優先
    NVIC_SetPriority(EXTI4_IRQn, 2);    // ３番目に優先

    // EXTI3 によるソフトウェア割り込みに対応する設定
    NVIC_SetVector(EXTI3_IRQn, (uint64_t)IsrCIC);
    NVIC_EnableIRQ(EXTI3_IRQn);

    // EXTI4 によるソフトウェア割り込みに対応する設定
    NVIC_SetVector(EXTI4_IRQn, (uint64_t)IsrFIR);
    NVIC_EnableIRQ(EXTI4_IRQn);

    // ADC に対する割り込みサービス･ルーチンの設定
    adc_.SetIntrVec(&AdcIsr);   // ここから AD 変換が開始

    printf("\r\nSDR を開始します\r\n");
    int k0 = -1;    // 以前の SW の値
    while (true)
        if (sw.read() != k0) k0 = Tune(lcd, sw);
}
